@@ -974,7 +974,7 @@ _hal_add_volume (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon,
 	 * changed 
 	 */
 	drive = _gnome_vfs_volume_monitor_find_drive_by_hal_udi (volume_monitor, libhal_volume_get_udi (hal_volume));
-	if (drive == NULL) {
+	if (drive == NULL && allowed_by_policy) {
 		drive = g_object_new (GNOME_VFS_TYPE_DRIVE, NULL);
 		if (libhal_volume_disc_has_audio (hal_volume)) {
 			drive->priv->activation_uri = g_strdup_printf ("cdda://%s", 
@@ -1054,8 +1054,8 @@ _hal_add_volume (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon,
 		name = _hal_volume_policy_get_display_name (volume_monitor_daemon, hal_drive, hal_volume);
 		vol->priv->display_name = _gnome_vfs_volume_monitor_uniquify_volume_name (volume_monitor, name);
 		g_free (name);
-		name = g_utf8_casefold (drive->priv->display_name, -1);
-		drive->priv->display_name_key = g_utf8_collate_key (name, -1);
+		name = g_utf8_casefold (vol->priv->display_name, -1);
+		vol->priv->display_name_key = g_utf8_collate_key (name, -1);
 		g_free (name);
 		vol->priv->icon = _hal_volume_policy_get_icon (volume_monitor_daemon, hal_drive, hal_volume);
 		vol->priv->is_user_visible = allowed_by_policy && 
@@ -1063,8 +1063,10 @@ _hal_add_volume (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon,
 		vol->priv->hal_udi = g_strdup (libhal_volume_get_udi (hal_volume));
 		vol->priv->hal_drive_udi = g_strdup (libhal_drive_get_udi (hal_drive));
 
-		vol->priv->drive = drive;
-		gnome_vfs_drive_add_mounted_volume_private (drive, vol);
+		if (drive) {
+			vol->priv->drive = drive;
+			gnome_vfs_drive_add_mounted_volume_private (drive, vol);
+		}
 
 #ifdef HAL_SHOW_DEBUG
 		g_debug ("Adding GnomeVFSVolume for device path %s", libhal_volume_get_device_file (hal_volume));
