@@ -28,7 +28,6 @@
 #include "gnome-vfs-mime.h"
 
 #include "gnome-vfs-configuration.h"
-#include "gnome-vfs-client.h"
 #include "gnome-vfs-method.h"
 #include "gnome-vfs-utils.h"
 #include "gnome-vfs-private-utils.h"
@@ -36,16 +35,20 @@
 #include "gnome-vfs-async-job-map.h"
 #include "gnome-vfs-job-queue.h"
 #include "gnome-vfs-volume-monitor-private.h"
+#include "gnome-vfs-module-callback-private.h"
 
 #include <errno.h>
-#include <bonobo-activation/bonobo-activation.h>
-#include <bonobo/bonobo-main.h>
 #include <glib/gmessages.h>
 #include <glib/gfileutils.h>
 #include <glib/gi18n-lib.h>
 #include <glib/gtypes.h>
 #include <glib/gstdio.h>
 
+#ifndef DBUS_API_SUBJECT_TO_CHANGE
+#define DBUS_API_SUBJECT_TO_CHANGE
+#endif
+#include <dbus/dbus-glib.h>
+
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <signal.h>
@@ -104,8 +107,10 @@ gboolean
 gnome_vfs_init (void)
 {
 	gboolean retval;
+	/*
 	char *bogus_argv[2] = { "dummy", NULL };
-
+	*/
+	
 	if (!ensure_dot_gnome_exists ()) {
 		return FALSE;
 	}
@@ -122,10 +127,9 @@ gnome_vfs_init (void)
 #endif   
 		gnome_vfs_thread_init ();
 
-		if (bonobo_activation_orb_get() == NULL) {
-			bonobo_activation_init (0, bogus_argv);
-		}
-		bonobo_init (NULL, bogus_argv);
+		dbus_g_thread_init ();
+ 		/* Make sure the type system is inited. */
+		g_type_init ();
 
 		_gnome_vfs_ssl_init ();
 
@@ -183,9 +187,6 @@ gnome_vfs_shutdown (void)
 #ifndef G_OS_WIN32
 	_gnome_vfs_volume_monitor_shutdown ();
 #endif
-	_gnome_vfs_client_shutdown ();
-	bonobo_debug_shutdown ();
-
 	_gnome_vfs_method_shutdown ();
 }
 
