@@ -22,6 +22,7 @@
 */
 
 #include <sys/time.h>
+#include <time.h>
 #include <string.h>
 #include <libgnomevfs/gnome-vfs-monitor.h>
 #include <libgnomevfs/gnome-vfs-monitor-private.h>
@@ -56,7 +57,7 @@ struct GnomeVFSMonitorCallbackData {
 	char *info_uri;
 	GnomeVFSMonitorEventType event_type;
 	CallbackState send_state;
-	guint32 send_at;
+	time_t send_at;
 };
 
 /* Number of seconds between consecutive events of the same type to the same file */
@@ -164,7 +165,7 @@ destroy_monitor_handle (GnomeVFSMonitorHandle *handle)
 	if (!res) {
 		g_warning ("gnome-vfs-monitor.c: A monitor handle was destroyed "
 			   "before it was added to the method hash table. This "
-			   "is a bug in the application and can cause crashed. "
+			   "is a bug in the application and can cause crashes. "
 			   "It is probably a race-condition.");
 	}
 
@@ -216,16 +217,14 @@ actually_dispatch_callback (gpointer data)
 	gchar *uri;
 	GList *l, *next;
 	GList *dispatch;
-	struct timeval tv;
-	guint32 now;
+	time_t now;
 
 	/* This function runs on the main loop, so it won't reenter,
 	 * although other threads may add stuff to the pending queue
 	 * while we don't have the lock
 	 */
 
-	gettimeofday (&tv, NULL);
-	now = tv.tv_sec;
+	time (&now);
 
 	G_LOCK (handle_hash);
 
@@ -331,7 +330,8 @@ send_uri_changes_now (GnomeVFSMonitorHandle *monitor_handle,
 	l = monitor_handle->pending_callbacks;
 	while (l != NULL) {
 		callback_data = l->data;
-		if (strcmp (callback_data->info_uri, uri) == 0) {
+		if (callback_data->send_state != CALLBACK_STATE_SENT &&
+		    strcmp (callback_data->info_uri, uri) == 0) {
 			callback_data->send_at = now;
 		}
 		l = l->next;
@@ -342,7 +342,7 @@ send_uri_changes_now (GnomeVFSMonitorHandle *monitor_handle,
 static guint32
 get_min_delay  (GList *list, gint32 now)
 {
-	guint32 min_send_at;
+	time_t min_send_at;
 	GnomeVFSMonitorCallbackData *callback_data;
 
 	min_send_at = G_MAXINT;
@@ -365,17 +365,25 @@ get_min_delay  (GList *list, gint32 now)
 }
 
 
-/* for modules to send callbacks to the app */
+/**
+ * gnome_vfs_monitor_callback:
+ * @method_handle: Method-specific monitor handle obtained through gnome_vfs_monitor_add().
+ * @info_uri: URI that triggered the callback.
+ * @event_type: The event obtained for @info_uri.
+ *
+ * gnome_vfs_monitor_callback() is used by #GnomeVFSMethods to indicate that a particular
+ * resource changed, and will issue the emission of the #GnomeVFSMonitorCallback registered
+ * using gnome_vfs_monitor_add().
+ **/
 void
 gnome_vfs_monitor_callback (GnomeVFSMethodHandle *method_handle,
-                            GnomeVFSURI *info_uri, /* GList of uris */
+                            GnomeVFSURI *info_uri,
                             GnomeVFSMonitorEventType event_type)
 {
 	GnomeVFSMonitorCallbackData *callback_data, *other_data, *last_data;
 	GnomeVFSMonitorHandle *monitor_handle;
 	char *uri;
-	struct timeval tv;
-	guint32 now;
+	time_t now;
 	guint32 delay;
 	GList *l;
 	DispatchData *ddata;
@@ -400,8 +408,7 @@ gnome_vfs_monitor_callback (GnomeVFSMethodHandle *method_handle,
 		return;
 	}
 	
-	gettimeofday (&tv, NULL);
-	now = tv.tv_sec;
+	time (&now);
 
 	uri = gnome_vfs_uri_to_string (info_uri, GNOME_VFS_URI_HIDE_NONE);
 
