@@ -46,10 +46,12 @@
 
 #ifdef HAVE_POSIX_ACL
 # include <acl/libacl.h>
+#define HAVE_ACL
 #endif
 
 #ifdef HAVE_SOLARIS_ACL
 # include <sys/acl.h>
+#define HAVE_ACL
 #endif
 
 #include "file-method-acl.h"
@@ -60,6 +62,8 @@
 #define CMD_PERM_COND_EXECUTE   8
 
 
+#ifdef HAVE_ACL
+
 static char *
 uid_to_string (uid_t uid) 
 {
@@ -805,6 +809,8 @@ GnomeVFSResult file_get_acl (const char       *path,
 #endif
 }
 
+#endif /* HAVE_ACL */
+
 GnomeVFSResult 
 file_set_acl (const char             *path,
 	      const GnomeVFSFileInfo *info,
