@@ -60,6 +60,10 @@
 #include <fam.h>
 #endif
 
+#ifdef HAVE_SELINUX
+#include <selinux/selinux.h>
+#endif
+
 #if defined(HAVE_LINUX_INOTIFY_H) || defined(HAVE_SYS_INOTIFY_H)
 #define USE_INOTIFY 1
 #include "inotify-helper.h"
@@ -74,6 +78,8 @@
 #include <sys/mount.h>
 #endif
 
+#include "file-method-acl.h"
+
 #ifdef G_OS_WIN32
 #include <windows.h>
 #endif
@@ -89,9 +95,11 @@ static FAMConnection *fam_connection = NULL;
 static gint fam_watch_id = 0;
 G_LOCK_DEFINE_STATIC (fam_connection);
 
+/* The first argument of this structure must be uri, so that it can be 
+   cast safely into a FileHandle. */
 typedef struct {
-	FAMRequest request;
 	GnomeVFSURI *uri;
+	FAMRequest request;
 	gboolean     cancelled;
 } FileMonitorHandle;
 #endif
@@ -720,6 +728,91 @@ read_link (const gchar *full_name)
 }
 #endif
 
+#ifdef HAVE_SELINUX
+/* convert a SELinux scurity context string to a g_malloc() compatible string */
+static char *sec_con2g_str(char *tmp)
+{
+        char *ret = tmp;
+  
+	if (tmp) {
+	         ret = g_strdup(tmp);
+		 freecon(tmp);
+	}
+
+	return ret;
+}
+#endif
+
+/* Get the SELinux security context */
+static int
+get_selinux_context (
+	GnomeVFSFileInfo *info,
+	const char *full_name,
+	GnomeVFSFileInfoOptions options)
+{
+#ifdef HAVE_SELINUX
+	if (is_selinux_enabled()) {
+	  
+		if ((options & GNOME_VFS_FILE_INFO_FOLLOW_LINKS) == 0
+			&& (info->type == GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK)) {
+
+			/* we are a symlink and aren't asked to follow -
+			 * return the type for a symlink */
+						
+			if (lgetfilecon_raw(full_name, &info->selinux_context) < 0)
+				return gnome_vfs_result_from_errno ();
+		} else {
+
+			if (getfilecon_raw(full_name, &info->selinux_context) < 0)
+				return gnome_vfs_result_from_errno ();
+		}
+
+		info->selinux_context = sec_con2g_str(info->selinux_context);
+		
+		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SELINUX_CONTEXT;
+		return GNOME_VFS_OK;
+	}
+#endif
+	return GNOME_VFS_OK;
+}
+
+/* Get the SELinux security context from handle */
+static int 
+get_selinux_context_from_handle (
+	GnomeVFSFileInfo *info,
+	FileHandle *handle)
+{
+#ifdef HAVE_SELINUX
+	if (is_selinux_enabled()) {
+		if (fgetfilecon_raw(handle->fd, &info->selinux_context) >= 0) 
+			return gnome_vfs_result_from_errno ();	
+
+		info->selinux_context = sec_con2g_str(info->selinux_context);
+		
+		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SELINUX_CONTEXT;
+		return GNOME_VFS_OK;
+	}	
+#endif
+	return GNOME_VFS_OK;
+}
+
+/* Set the SELinux security context */
+static int 
+set_selinux_context (
+	const GnomeVFSFileInfo *info,
+	const char *full_name) 
+{
+#ifdef HAVE_SELINUX
+	if (is_selinux_enabled()) {
+		if (setfilecon_raw(full_name, info->selinux_context) < 0)
+			return gnome_vfs_result_from_errno ();
+
+		return GNOME_VFS_OK;
+	}
+#endif
+	return GNOME_VFS_OK;
+}
+
 static void
 get_access_info (GnomeVFSFileInfo *file_info,
 		 const gchar *full_name)
@@ -729,7 +822,6 @@ get_access_info (GnomeVFSFileInfo *file_info,
       * of the access calls
       */
 #ifdef G_OS_WIN32
-	file_info->permissions = 0;
 	if (g_access (full_name, R_OK) == 0) {
 		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
 	}
@@ -745,17 +837,16 @@ get_access_info (GnomeVFSFileInfo *file_info,
 	 * calls needed
 	 */
 	if (g_access (full_name, R_OK|W_OK) == 0) {
-		file_info->permissions = GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_WRITABLE;
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_WRITABLE;
 		if (g_access (full_name, X_OK) == 0) {
 			file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
 		}
 	} else if (g_access (full_name, R_OK|X_OK) == 0) {
-		file_info->permissions = GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+		file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE | GNOME_VFS_PERM_ACCESS_EXECUTABLE;
 	} else {
 		if (g_access (full_name, R_OK) == 0) {
-			file_info->permissions = GNOME_VFS_PERM_ACCESS_READABLE;
+			file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
 		} else {
-			file_info->permissions = 0;
 			if (g_access (full_name, W_OK) == 0) {
 				file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
 			}
@@ -776,7 +867,6 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 	       struct stat *statptr)
 {
 	struct stat statbuf;
-	gboolean followed_symlink;
 #ifndef G_OS_WIN32
 	gboolean is_symlink;
 	char *link_file_path;
@@ -786,8 +876,6 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 #endif
 	gboolean recursive;
 	
-	followed_symlink = FALSE;
-	
 	recursive = FALSE;
 
 	GNOME_VFS_FILE_INFO_SET_LOCAL (file_info, TRUE);
@@ -818,7 +906,6 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 			}
 		}
 		GNOME_VFS_FILE_INFO_SET_SYMLINK (file_info, TRUE);
-		followed_symlink = TRUE;
 	}
 #endif
 	gnome_vfs_stat_to_file_info (file_info, statptr);
@@ -1024,6 +1111,12 @@ do_read_directory (GnomeVFSMethod *method,
 	if (handle->options & GNOME_VFS_FILE_INFO_NAME_ONLY) {
 		return GNOME_VFS_OK;
 	}
+
+	if (handle->options & GNOME_VFS_FILE_INFO_GET_SELINUX_CONTEXT) {
+
+		/* Attempt to get selinux contet, ignore error (see below) */
+		get_selinux_context(file_info, full_name, handle->options);
+	}
 		
 	if (get_stat_info (file_info, full_name, handle->options, &statbuf) != GNOME_VFS_OK) {
 		/* Return OK - this should not terminate the directory iteration
@@ -1041,6 +1134,10 @@ do_read_directory (GnomeVFSMethod *method,
 		get_mime_type (file_info, full_name, handle->options, &statbuf);
 	}
 
+	if (handle->options & GNOME_VFS_FILE_INFO_GET_ACL) {
+		file_get_acl (full_name, file_info, &statbuf, context);
+	}
+
 	return GNOME_VFS_OK;
 }
 
@@ -1070,6 +1167,14 @@ do_get_file_info (GnomeVFSMethod *method,
 		return result;
 	}
 
+	if (options & GNOME_VFS_FILE_INFO_GET_SELINUX_CONTEXT) {
+		result = get_selinux_context (file_info, full_name, options);
+		if (result != GNOME_VFS_OK) { 
+			g_free (full_name);
+			return result;
+		}
+	} 
+
 	if (options & GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) {
 		get_access_info (file_info, full_name);
 	}
@@ -1077,6 +1182,10 @@ do_get_file_info (GnomeVFSMethod *method,
 	if (options & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
 		get_mime_type (file_info, full_name, options, &statbuf);
 	}
+	
+	if (options & GNOME_VFS_FILE_INFO_GET_ACL) {
+		file_get_acl (full_name, file_info, &statbuf, context);	
+	}
 
 	g_free (full_name);
 
@@ -1114,10 +1223,22 @@ do_get_file_info_from_handle (GnomeVFSMethod *method,
 		return result;
 	}
 
+	if (options & GNOME_VFS_FILE_INFO_GET_SELINUX_CONTEXT) {
+		result = get_selinux_context_from_handle (file_info, file_handle);
+		if (result != GNOME_VFS_OK) {
+			g_free (full_name);
+			return result;
+		}
+	}
+
 	if (options & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
 		get_mime_type (file_info, full_name, options, &statbuf);
 	}
 
+	if (options & GNOME_VFS_FILE_INFO_GET_ACL) {
+		file_get_acl (full_name, file_info, &statbuf, context);	
+	}
+
 	g_free (full_name);
 
 	return GNOME_VFS_OK;
@@ -1864,9 +1985,6 @@ do_find_directory (GnomeVFSMethod *method,
 			/* This volume does not contain our home, we have to find/create the Trash
 			 * elsewhere on the volume. Use a heuristic to find a good place.
 			 */
-			FindByDeviceIDParameters tmp;
-			tmp.device_id = near_item_stat.st_dev;
-
 			if (gnome_vfs_context_check_cancellation (context))
 				return GNOME_VFS_ERROR_CANCELLED;
 
@@ -2203,9 +2321,10 @@ do_set_file_info (GnomeVFSMethod *method,
 		new_name = g_build_filename (dir, info->name, NULL);
 
 		result = rename_helper (full_name, new_name, FALSE, context);
-
+		
 		g_free (dir);
-		g_free (new_name);
+		g_free (full_name);
+		full_name = new_name;
 
 		if (result != GNOME_VFS_OK) {
 			g_free (full_name);
@@ -2213,6 +2332,14 @@ do_set_file_info (GnomeVFSMethod *method,
 		}
 	}
 
+	if (mask & GNOME_VFS_SET_FILE_INFO_SELINUX_CONTEXT) {
+		GnomeVFSResult result = set_selinux_context(info, full_name);
+		if (result < 0) {
+			g_free(full_name);
+			return result;
+		}
+	}
+
 	if (gnome_vfs_context_check_cancellation (context)) {
 		g_free (full_name);
 		return GNOME_VFS_ERROR_CANCELLED;
@@ -2240,6 +2367,7 @@ do_set_file_info (GnomeVFSMethod *method,
 		g_warning ("Not implemented: GNOME_VFS_SET_FILE_INFO_OWNER");
 #endif
 	}
+	
 	if (gnome_vfs_context_check_cancellation (context)) {
 		g_free (full_name);
 		return GNOME_VFS_ERROR_CANCELLED;
@@ -2257,6 +2385,22 @@ do_set_file_info (GnomeVFSMethod *method,
 		}
 	}
 
+	if (gnome_vfs_context_check_cancellation (context)) {
+		g_free (full_name);
+		return GNOME_VFS_ERROR_CANCELLED;
+	}
+	
+	if (mask & GNOME_VFS_SET_FILE_INFO_ACL) {
+		GnomeVFSResult result;
+
+		result = file_set_acl (full_name, info, context);
+		if (result != GNOME_VFS_OK) {
+			g_free (full_name);
+			return result;	
+		}
+	}
+	
+
 	g_free (full_name);
 
 	return GNOME_VFS_OK;
@@ -2473,7 +2617,8 @@ do_monitor_add (GnomeVFSMethod *method,
 		GnomeVFSMonitorType monitor_type)
 {
 #ifdef USE_INOTIFY
-	if (ih_startup ()) {
+	/* For remote files, always prefer FAM */
+	if (do_is_local (method, uri) && ih_startup ()) {
 		return inotify_monitor_add (method, method_handle_return, uri, monitor_type);
 	}
 #endif
@@ -2537,7 +2682,12 @@ do_monitor_cancel (GnomeVFSMethod *method,
 		   GnomeVFSMethodHandle *method_handle)
 {
 #ifdef USE_INOTIFY
-	if (ih_startup ()) {
+	FileHandle *file_handle;
+	file_handle = (FileHandle *) method_handle;
+	/* It is possible to call file_handle->uri as it can be either 
+	 * a FileMonitorHandle or a ih_sub_t, and both structures also
+	 * have a GnomeVFSURI as first element. */
+	if (do_is_local (method, file_handle->uri) && ih_startup ()) {
 		return inotify_monitor_cancel (method, method_handle);
 	}
 #endif
