@@ -283,7 +283,7 @@ ftp_response_to_vfs_result (FtpConnection *conn)
 }
 
 #ifdef HAVE_GSSAPI
-static char *radixN =
+static const char *radixN =
 	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static char pad = '=';
@@ -857,6 +857,13 @@ save_authn_info (GnomeVFSURI *uri,
         g_free (in_args.uri);
 }
 
+/*
+ * Returns FALSE if callback was not handled (*aborted
+ *   will be FALSE, user will be "anonymous")
+ *
+ * Returns TRUE if callback invocation succeeded, *aborted
+ *   will be set to TRUE only if the user didn't cancel
+ */
 static gboolean
 query_user_for_authn_info (GnomeVFSURI *uri, 
 			   char **user, char **pass, char **keyring,
@@ -867,7 +874,7 @@ query_user_for_authn_info (GnomeVFSURI *uri,
         GnomeVFSModuleCallbackFullAuthenticationOut out_args;
 	gboolean ret;
 
-	ret = FALSE;
+	ret = *aborted = FALSE;
 	
 	memset (&in_args, 0, sizeof (in_args));
         memset (&out_args, 0, sizeof (out_args));
@@ -895,7 +902,6 @@ query_user_for_authn_info (GnomeVFSURI *uri,
                                                 &out_args, sizeof (out_args));
 
 	if (!ret) {
-		ret = TRUE;
 		/* No callback, try anon login */
 		*user = g_strdup ("anonymous");
 		*pass = g_strdup ("nobody@gnome.org");
@@ -934,7 +940,7 @@ query_user_for_authn_info (GnomeVFSURI *uri,
        	g_free (out_args.password);
        	g_free (out_args.keyring);
 
-	return ret && !out_args.abort_auth;
+	return ret;
 } 
 
 static gboolean
@@ -1005,9 +1011,7 @@ ftp_kerberos_login (FtpConnection *conn,
 	char *encoded_token;
 	int len;
 	GnomeVFSResult result;
-	gboolean auth_ok;
 
-	auth_ok = FALSE;
 	result = do_basic_command (conn, "AUTH GSSAPI", cancellation);
 	if (result != GNOME_VFS_OK) {
 		return result;
@@ -1390,7 +1394,7 @@ ftp_connection_create (FtpConnectionPool *pool,
 								 &aborted, !uri_has_username); 
 				G_LOCK (connection_pools);
 				pool->num_connections--;
-				if (!ret) {
+				if (aborted) {
 					gnome_vfs_uri_unref (conn->uri);
 					g_string_free (conn->response_buffer, TRUE);
 					g_free (conn);
@@ -1398,11 +1402,7 @@ ftp_connection_create (FtpConnectionPool *pool,
 					g_free (pass);
 					g_free (keyring);
 
-					if (aborted) {
-						return GNOME_VFS_ERROR_CANCELLED;
-					} else {
-						return GNOME_VFS_ERROR_LOGIN_FAILED;
-					}
+					return GNOME_VFS_ERROR_CANCELLED;
 				}
 				g_string_free (conn->response_buffer, TRUE);
                         	conn->response_buffer = g_string_new ("");
@@ -1412,7 +1412,9 @@ ftp_connection_create (FtpConnectionPool *pool,
 				if (result == GNOME_VFS_OK) {
 					break;
 				}
-				if (result != GNOME_VFS_ERROR_LOGIN_FAILED) {
+				if (result != GNOME_VFS_ERROR_LOGIN_FAILED ||
+				    !ret /* if callback was not handled, and anonymous
+					    login failed, don't run into endless loop */) {
 					gnome_vfs_uri_unref (conn->uri);
 					g_string_free (conn->response_buffer, TRUE);
 					g_free (conn);
@@ -1864,7 +1866,52 @@ do_create (GnomeVFSMethod *method,
 	   guint perm,
 	   GnomeVFSContext *context)
 {
-	return do_open(method, method_handle, uri, mode, context);
+	GnomeVFSResult result;
+	FtpConnection *conn;
+	gchar *chmod_command;
+
+	if ((mode & GNOME_VFS_OPEN_READ) && (mode & GNOME_VFS_OPEN_WRITE)) {
+		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
+	}
+
+	if (!(mode & GNOME_VFS_OPEN_READ) && !(mode & GNOME_VFS_OPEN_WRITE)) {
+		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
+	}
+
+	result = ftp_connection_acquire (uri, &conn, context);
+	if (result != GNOME_VFS_OK)
+		return result;
+
+	/* Exclusive set: we have to fail if the file exists */
+	if (exclusive) {
+		conn->operation = FTP_READ;
+		result = do_path_transfer_command (conn, "RETR", uri, context);
+
+		/* File exists or an error occurred */
+		if (result != GNOME_VFS_ERROR_NOT_FOUND) {
+			ftp_connection_release (conn, TRUE);
+
+			/* Return the error if there is one, otherwise the file exists */
+			return (result == GNOME_VFS_OK) ? GNOME_VFS_ERROR_FILE_EXISTS
+		                                        : result;
+		}
+	}
+
+	result = do_open(method, method_handle, uri, mode, context);
+	if (result != GNOME_VFS_OK) {
+		ftp_connection_release (conn, TRUE);
+		return result;
+	}
+
+	/* This is a not-standard command and FTP server may not support it,
+	 * so eventual errors will be ignored */
+	chmod_command = g_strdup_printf("SITE CHMOD %o", perm);
+	do_path_command (conn, chmod_command, uri, get_cancellation (context));
+
+	g_free (chmod_command);
+	ftp_connection_release (conn, TRUE);
+
+	return result;
 }
 
 static GnomeVFSResult 
@@ -2600,7 +2647,7 @@ do_read_directory (GnomeVFSMethod *method,
 					break;
 				}
 				
-				symlink_name = g_strdup (symlink_info->symlink_name);
+				symlink_name = gnome_vfs_escape_path_string (symlink_info->symlink_name);
 				gnome_vfs_file_info_clear (symlink_info);
 				
 				link_uri = gnome_vfs_uri_resolve_relative (uri, symlink_name);
@@ -2691,6 +2738,15 @@ do_make_directory (GnomeVFSMethod *method,
 	GnomeVFSResult result;
 	gchar *chmod_command;
 
+	result = do_path_command_completely ("CWD", uri, context,
+			/* If the directory doesn't exist @result will be !GNOME_VFS_OK,
+			* in this case we don't return and proceed with MKD */
+			!GNOME_VFS_OK);
+
+	if (result == GNOME_VFS_OK) {
+		return GNOME_VFS_ERROR_FILE_EXISTS;
+	}
+
 	result = do_path_command_completely ("MKD", uri, context, 
 		GNOME_VFS_ERROR_ACCESS_DENIED);
 
